/**
 * 
 */
package de.tu_darmstadt.tk.SmartHomeNetworkSim.core.anomalydetectors;

import java.io.File;
import java.io.IOException;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map.Entry;

import de.tu_darmstadt.tk.SmartHomeNetworkSim.core.Link;
import de.tu_darmstadt.tk.SmartHomeNetworkSim.core.Packet;
import de.tu_darmstadt.tk.SmartHomeNetworkSim.core.PacketSniffer;
import weka.classifiers.AbstractClassifier;
import weka.core.Attribute;
import weka.core.DenseInstance;
import weka.core.Instance;
import weka.core.Instances;
import weka.core.converters.ArffSaver;

/**
 * @author Camille Peto
 *
 */
public abstract class BaseAnomalyDetector implements PacketSniffer {

	/**
	 * True, if instances should be used for training
	 */
	protected boolean training = true;
	
	/**
	 * Attributes which should be taken into account
	 */
	protected ArrayList<Attribute> atts = new ArrayList<Attribute>();
	
	/**
	 * Collected Packets
	 */
	protected Instances dataset;
	
	//Evaluation variables
	protected long tp = 0;
	protected long fp = 0;
	protected long fn = 0;
	protected long tn = 0;
	
	public BaseAnomalyDetector() {
		atts.add(new Attribute("Link-Name", false));
		atts.add(new Attribute("Source-Port-number", false));
		atts.add(new Attribute("Destination-Port-number", false));
		atts.add(new Attribute("Packets-per-simulationstep", false));
		atts.add(new Attribute("Average Value", false));
		// Initialize data set
		dataset = new Instances("Directions", atts, 100000);
	}
	//create Instances out of packets and trains/classifies them afterwards
	@Override
	public void processPackets(HashMap<Link, LinkedList<Packet>> packets) {
		for (Iterator<Entry<Link, LinkedList<Packet>>> it = packets.entrySet().iterator(); it.hasNext();) {
			Entry<Link, LinkedList<Packet>> entry = it.next();
			Link l = entry.getKey();
			
			float[][] directions = new float[l.getPackets().size()][4];
			for (Iterator<Packet> itPacket = entry.getValue().iterator(); itPacket.hasNext();) {
				Packet packet = (Packet) itPacket.next();
				for(int i = 0; i < directions.length; i++) {
					int j = 0;
					boolean sameSource = false;
					if(directions[i][j] != packet.getSource().getPortNumber() && directions[i][j] == 0) {
						directions[i][j] = packet.getSource().getPortNumber();
						sameSource = true;
					}
					else if(directions[i][j] == packet.getSource().getPortNumber()) sameSource = true;
					
					j++;
					
					boolean sameDestination = false;
					if(directions[i][j] != packet.getDestination().getPortNumber() && directions[i][j] == 0) {
						directions[i][j] = packet.getDestination().getPortNumber();
						sameDestination = true;
					}
					else if(directions[i][j] == packet.getDestination().getPortNumber()) sameDestination = true;
					
					j++;
					
					if(sameSource && sameDestination) {
						directions[i][j] = directions[i][j] + 1;
						j++;
						directions[i][j] = (directions[i][j] + packet.getSource().getOwner().getValue()); //combined value
						break;
					}
					
					
				}
				

				
			}
			
			for(int i = 0; i < directions.length; i ++) {
				if(training)
					try {
						training(l, directions[i]);
					} catch (Exception e) {
						e.printStackTrace();
					}
				else {
					classify(l, directions[i]);
					}
				
			}
			
			if(!training && !getAlgoName().contentEquals("K-Means")) {
				System.out.println("Metrics for " + getAlgoName() + ": TP = " + tp + ", FP = " + fp + ", FN = " + fn + ", TN = " + tn);
			}
		}
		
	}
	
	public void classify(Link l, float[] direction) {
		try {
			Instance instance = direction2Instance(l, direction);
			if(instance == null) {
				return;
			}
			
			boolean isClassifiedAanomaly = classifyInstance(instance);
			boolean isActualAnomaly = direction[0] > 6 || direction[0] < 0 || direction[1] > 6 || direction[1] < 0 || //unknown portnumber
									direction[2] > 100 || direction[2] < 0 || //number of packages is too high/low
									(direction[3]/direction[2]) > 40 || (direction[3]/direction[2]) < 0; //Average Value tested
			if(isClassifiedAanomaly) {
				if(isActualAnomaly) {
					tp++;
					
				}else {
					System.out.println("isActualAnomaly; " + isActualAnomaly);
					fp++;
				}
			}else {
				if(isActualAnomaly) {
					System.out.println("isActualAnomaly; " + isActualAnomaly);
					fn++;
					
				}else {
					tn++;
				}
			}
			
			} catch (Exception e) {
					System.out.println("Error classifying instance");
					e.printStackTrace();
			}
	}
	
	/**
	 * Collect the packets
	 * 
	 * @param packets packets to be learned
	 */
	protected void training(Link l, float[] direction) {
		Instance instance = direction2Instance(l, direction);
		if(instance == null) {
			return;
		}
		dataset.add(instance);
	}
	
	private Instance direction2Instance(Link link, float[] direction) {
		
			DenseInstance instance = new DenseInstance(dataset.numAttributes());
			instance.setDataset(dataset);
			
			if(direction[0] == 0 && direction[1] == 0) {
				return null;
			}
			// link (currently always 1)
			instance.setValue(0, 1);
			
			//Source
			instance.setValue(1, direction[0]);
			
			//Destination
			instance.setValue(2, direction[1]);
			
			//NumOfPackets
			instance.setValue(3, direction[2]);
			
			//Average Value
			instance.setValue(4, direction[3]/direction[2]);
			
			
			return instance;
	}

	@Override
	public void setMode(boolean testing) {
		training = !testing;
		if(testing) {
			try {
				finishDataCollection();
			} catch (Exception e) {
				System.out.println("Clustering failed");
				e.printStackTrace();
			}	
		}
	}

	@Override
	public boolean getMode() {
		return !training;
	}

	
	/**
	 * Finishes the collection and trains the clusterer on the collected packets
	 * 
	 * @throws Exception
	 */
	private void finishDataCollection() throws Exception{
		System.out.println("Number of instances: " + dataset.numInstances());
		
		trainModel(dataset);
	}
	
	/**
	 * Train the model using the given instances
	 * @param instances training set, which should be learned
	 */
	public abstract void trainModel(Instances instances);
	
	public abstract boolean classifyInstance(Instance instance) throws Exception;
	
	public abstract void saveModel() throws Exception;
	
	public abstract String getAlgoName();
}
